import copy
import math

import torch
import numpy as np

import vqt

    
import pytest
from hypothesis import given, strategies as st, settings, Verbosity, assume, example, seed, HealthCheck
import torch
import numpy as np
import scipy.signal

params_vqt = {
    'Fs_sample': 1000, 
    'Q_lowF': 3, 
    'Q_highF': 20, 
    'F_min': 10, 
    'F_max': 400, 
    'n_freq_bins': 55, 
    'win_size': None,
    'symmetry': 'center',
    'taper_asymmetric': True,
    'downsample_factor': 4,
    'padding': 'valid',
    'take_abs': True,
    'filters': None,
    'verbose': False,
    'plot_pref': False,
}


# 1. Test Zero Signal Transformation
def test_zero_signal_transformation():
    params = copy.deepcopy(params_vqt) 
    v = vqt.VQT(**params)  # Create a new instance for each test case
    input_signal = torch.zeros(1024, dtype=torch.float32)  # A zero signal of length 1024
    output = v(input_signal)
    assert torch.all(output == 0), "VQT output for a zero signal should be zero across all frequency bins"

# 2. Test Impulse Signal Transformation
def test_impulse_signal_transformation():
    params = copy.deepcopy(params_vqt) 
    v = vqt.VQT(**params)  # Create a new instance for each test case
    input_signal = torch.zeros(1024, dtype=torch.float32)
    input_signal[512] = 1  # Set the middle point to 1, creating an impulse
    output = v(input_signal)
    assert not torch.all(output == 0), "VQT output for an impulse signal should not be zero"
    assert output.shape[1] == v.filters.shape[0], "VQT output should have the same number of frequency bins as filters"

# 3. Test Sinusoidal Signal Transformation
@given(frequency=st.floats(
    # min_value=copy.deepcopy(params_vqt['F_min']), 
    # min_value=params_vqt['F_min'], 
    min_value=10,
    max_value=400,
    # max_value=copy.deepcopy(params_vqt['F_max']),
))
@settings(deadline=500)
def test_peak_in_spectrogram_at_sine_wave_frequency(
    frequency,
):
    """Test to verify a peak in the spectrogram at a specific sine wave frequency."""
    params = copy.deepcopy(params_vqt) 
    v = vqt.VQT(**params)  # Create a new instance for each test case
    # Generate the sine wave signal
    t = torch.arange(0, 1.0, 1/params['Fs_sample'], dtype=torch.float32)
    input_signal = torch.sin(2 * np.pi * frequency * t)
    # Apply the VQT to this sine wave
    spectrogram = v(input_signal)
    freqs = v.get_freqs()
    # Convert freqs to a tensor for easier handling
    freqs_tensor = torch.as_tensor(freqs, dtype=torch.float32)
    # Locate the peak in the spectrogram
    peak_index = torch.argmax(spectrogram[0], dim=0)  # Assuming the output shape is (n_channels, n_freq_bins, time_bins) 
    assert torch.all(peak_index == peak_index[0]), "Expected a single peak in the spectrogram"
    peak_index = peak_index[0]
    # Find the frequency corresponding to the peak
    peak_frequency = freqs_tensor[peak_index]  # Taking the first peak if multiple time bins
    # Frequency nearest within the discrete frequency steps
    idx_nearest = torch.argmin(torch.abs(freqs_tensor - frequency))
    frequency_nearest = freqs_tensor[idx_nearest]
    # print(frequency, frequency_nearest.item(), peak_frequency.item(), peak_index.item(), idx_nearest.item())
    # Verify that the peak corresponds to the sine wave's frequency
    # assert torch.abs(peak_frequency - frequency_nearest) < frequency_tolerance, \
    #     f"Expected a peak at {frequency_nearest} Hz, found one at {peak_frequency.item()} Hz within a tolerance of {frequency_tolerance} Hz."
    assert np.abs(peak_index.item() - idx_nearest.item()) <= 1, \
        f"Expected a peak at index: {idx_nearest.item()} and frequency: {frequency_nearest.item()} Hz, found one at index: {peak_index.item()} and frequency: {peak_frequency.item()} Hz within a tolerance of 2 indices."
        
# 4. Test Constant Signal Transformation
def test_constant_signal_transformation():
    params = copy.deepcopy(params_vqt) 
    v = vqt.VQT(**params)  # Create a new instance for each test case
    input_signal = torch.ones(1024, dtype=torch.float32)  # A constant signal
    output = v(input_signal)
    # In a constant signal, the energy should be concentrated at the lowest frequency bin
    assert torch.all(output[0] > output[1:]), "VQT output for a constant signal should peak at the lowest frequency bin"

# 5. Test various params
@settings(max_examples=1000, deadline=None, suppress_health_check=(HealthCheck.too_slow,))
@given(
    Fs_sample=st.floats(min_value=1, max_value=10000),
    Q_lowF=st.floats(min_value=1, max_value=100),
    Q_highF=st.floats(min_value=1, max_value=100),
    F_min=st.floats(min_value=1, max_value=1000),
    F_max=st.floats(min_value=1, max_value=1000),
    n_freq_bins=st.integers(min_value=1, max_value=100),
    win_size=st.integers(min_value=1, max_value=1000) | st.none(),
    symmetry=st.sampled_from(['center', 'left', 'right']),
    taper_asymmetric=st.booleans(),
    downsample_factor=st.integers(min_value=1, max_value=100),
    padding=st.sampled_from(['valid', 'same']),
    # filters=st.none() | st.sampled_from([None]),
    # plot_pref=st.booleans(),

    n_channels=st.integers(min_value=1, max_value=100),
    n_dim=st.integers(min_value=1, max_value=2),
)
def test_vqt_params(
    Fs_sample,
    Q_lowF,
    Q_highF,
    F_min,
    F_max,
    n_freq_bins,
    win_size,
    symmetry,
    taper_asymmetric,
    downsample_factor,
    padding,
    # take_abs,
    # filters,
    # plot_pref,

    n_channels,
    n_dim,
):
    ## Clip win_size to be at most 1001 and just set it 
    if win_size is not None:
        win_size = min(win_size, 1001)
    else:
        win_size = min(int(math.ceil(Q_lowF * (Fs_sample / F_min))), 1001)
    params = {
        'Fs_sample': Fs_sample,
        'Q_lowF': Q_lowF,
        'Q_highF': Q_highF,
        'F_min': F_min,
        'F_max': F_max,
        'n_freq_bins': n_freq_bins,
        'win_size': win_size,
        'symmetry': symmetry,
        'taper_asymmetric': taper_asymmetric,
        'downsample_factor': downsample_factor,
        'padding': padding,
        # 'take_abs': take_abs,
        # 'filters': filters,
        # 'plot_pref': plot_pref,
    }
    params.update({k: v for k, v in params_vqt.items() if k not in params})
    v = vqt.VQT(**params)

    # Make signal
    len_signal = params['win_size'] if params['win_size'] is not None else int(params['Q_lowF'] * params['Fs_sample'])
    input_signal = torch.rand(n_channels, len_signal + np.random.randint(1, 1000), dtype=torch.float32)
    if n_dim == 1:
        input_signal = input_signal[0]
    # Apply the VQT to this signal
    output = v(input_signal)
    
    # Perform validations on the output
    assert output is not None, "Output should not be None"
    # Check output shape
    assert output.shape[1] == params['n_freq_bins'], "VQT output shape does not match the number of frequency bins"
    # Check all output is real if take_abs is True
    if params['take_abs']:
        assert torch.all(torch.isreal(output)), "VQT output should be real if take_abs is True"


def test_vqt_filters():
    params = copy.deepcopy(params_vqt) 
    # Generate the filters
    filters, freqs, wins = vqt.helpers.make_VQT_filters(
        Fs_sample=params['Fs_sample'],
        Q_lowF=params['Q_lowF'],
        Q_highF=params['Q_highF'],
        F_min=params['F_min'],
        F_max=params['F_max'],
        n_freq_bins=params['n_freq_bins'],
        win_size=params['win_size'],
        symmetry=params['symmetry'],
        taper_asymmetric=params['taper_asymmetric'],
        plot_pref=params['plot_pref']
    )
    filters, freqs, wins = filters.numpy(), freqs.numpy(), wins.numpy()

    ## Filters, freqs, and wins should all have the following properties:
    ### They are not all zeros, not all ones, and do not contain NaNs or infinities
    ### Filters should additionally have shape (n_freq_bins, win_size)
    if params['win_size'] is not None:
        assert filters.shape == (params['n_freq_bins'], params['win_size']), "VQT filters have an unexpected shape"
    else:
        assert filters.shape[0] == params['n_freq_bins'], "VQT filters have an unexpected shape"
    fns = {
        'not_all_zeros': lambda x: ~np.all(x == 0),
        'not_all_ones': lambda x: ~np.all(x == 1),
        'no_nans': lambda x: ~np.any(np.isnan(x)),
        'no_infs': lambda x: ~np.any(np.isinf(x)),
    }
    args = {
        'filters': filters,
        'freqs': freqs,
        'wins': wins,
    }
    for fn_name, fn in fns.items():
        for arg_name, arg in args.items():
            assert fn(arg), f"Test failed: arg: {arg_name} failed test: {fn_name}"
    
    ## Check that the filters are complex
    assert np.iscomplexobj(filters), "VQT filters are not complex"


#####################
## Other functions ##
#####################


# Test Hilbert Transform Equivalence
def test_hilbert_transform_equivalence():
    params = copy.deepcopy(params_vqt) 
    # Generate a random signal
    input_signal = np.random.rand(1024).astype(np.float64)  ## Implementations for fft are different between numpy and torch for float32
    # Use the VQT class's Hilbert transform
    hilbert_vqt = vqt.helpers.torch_hilbert(torch.as_tensor(input_signal)).numpy()
    # Compare with scipy's Hilbert transform
    hilbert_scipy = scipy.signal.hilbert(input_signal)
    # Check if they are approximately equal
    assert np.allclose(hilbert_vqt, hilbert_scipy, rtol=1e-6), "Hilbert transforms do not match"
 

# Test convolution functions
def test_conv_accuracy():
    conditions = {
        'x': [
            ('ones_6', np.ones(6).astype(np.float32)),
            ('zeros_1center', np.array([0, 0, 0, 1, 0, 0]).astype(np.float32)),
        ],
        'k': [
            ('ones_3', np.ones(3).astype(np.float32)),
            ('ones_4', np.ones(4).astype(np.float32)),
            ('mid_1', np.array([0, 1, 0]).astype(np.float32)),
            ('mid_1_off', np.array([0, 0, 1, 0]).astype(np.float32)),
        ],
        'padding': ['valid', 'same'],
        'real': [False, True],
        'take_abs': [False, True],
        'fft': [False, True],
    }

    results = {}  # Dictionary to store results

    # Iterate through all combinations of conditions
    for x_name, x in conditions['x']:
        for k_name, k in conditions['k']:
            for padding in conditions['padding']:
                for take_abs in conditions['take_abs']:
                    for real in conditions['real']:
                        k_tu = scipy.signal.hilbert(k) if not real else copy.deepcopy(k)
                        for fft in conditions['fft']:
                            y_vqt = vqt.vqt.convolve(
                                torch.as_tensor(x),
                                torch.as_tensor(k_tu),
                                padding=padding, 
                                fast_length=True,
                                take_abs=take_abs,
                                fft_conv=fft
                            ).numpy()[0][0]
                            y_np = np.convolve(x, k_tu, mode=padding)
                            y_np = np.abs(y_np) if take_abs else y_np
                            key = (x_name, k_name, padding, take_abs, real, fft)
                            results[key] = {'y_vqt': np.round(y_vqt, 2), 'y_np': np.round(y_np, 2)}
            
        # Assertions and comparisons
        for key in results:
            x_name, k_name, padding, take_abs, real, fft = key
            # Compare FFT True vs FFT False results for each condition except fft flag
            if fft:  # Only check when fft is True, comparing it to fft False
                key_no_fft = (x_name, k_name, padding, take_abs, real, False)
                close = np.allclose(results[key]['y_vqt'], results[key_no_fft]['y_vqt'], rtol=1e-5)
                if not close:
                    print(f"Comparing {key} with FFT False")
                    print(f"result: {close}")
                    print(results[key]['y_vqt'])
                    print(results[key_no_fft]['y_vqt'])
                    print()
                assert close, f"Convolution result for {key} does not match FFT False"
            
            # Compare to NumPy's result when padding is 'valid'
            if padding == 'valid':
                close = np.allclose(results[key]['y_vqt'], results[key]['y_np'], rtol=1e-5)
                if not close:
                    print(f"Comparing {key} with NumPy's convolution result")
                    print(f"result: {close}")
                    print(results[key]['y_vqt'])
                    print(results[key]['y_np'])
                    print()
                assert close, f"Convolution result for {key} does not match NumPy's"

            # Compare to known outputs when when padding is 'same'
            if padding == 'same' and real == True:
                if x_name == 'ones_6' and k_name == 'ones_4':
                    close = np.allclose(results[key]['y_vqt'], np.array([3, 4, 4, 4, 3, 2]), rtol=1e-5)
                    if not close:
                        print(f"Comparing {key} with known output")
                        print(f"result: {close}")
                        print(results[key]['y_vqt'])
                        print(np.array([3, 4, 4, 4, 3, 2]))
                    assert close, f"Convolution result for {key} does not match expected output"


import math
from typing import Union, List, Tuple, Dict, Any, Optional, Sequence, Iterable

import torch
import numpy as np
import scipy.signal
import matplotlib.pyplot as plt

def make_batches(
    iterable, 
    batch_size=None, 
    num_batches=None, 
    min_batch_size=0, 
    return_idx=False, 
    length=None,
    idx_start=0,
):
    """
    Make batches of data or any other iterable.
    RH 2021

    Args:
        iterable (iterable):
            iterable to be batched
        batch_size (int):
            size of each batch
            if None, then batch_size based on num_batches
        num_batches (int):
            number of batches to make
        min_batch_size (int):
            minimum size of each batch
        return_idx (bool):
            whether to return the slice indices of the batches.
            output will be [start, end] idx
        length (int):
            length of the iterable.
            if None, then length is len(iterable)
            This is useful if you want to make batches of 
             something that doesn't have a __len__ method.
        idx_start (int):
            starting index of the iterable.
    
    Returns:
        output (iterable):
            batches of iterable
    """

    if length is None:
        l = len(iterable)
    else:
        l = length
    
    if batch_size is None:
        batch_size = int(math.ceil(l / num_batches))
    
    for start in range(idx_start, l, batch_size):
        end = min(start + batch_size, l)
        if (end-start) < min_batch_size:
            break
        else:
            if return_idx:
                yield iterable[start:end], [start, end]
            else:
                yield iterable[start:end]


def gaussian(x=None, mu=0, sig=1):
    '''
    A gaussian function (normalized similarly to scipy's function)
    RH 2021
    
    Args:
        x (np.ndarray): 1-D array of the x-axis of the kernel
        mu (float): center position on x-axis
        sig (float): standard deviation (sigma) of gaussian
        
    Returns:
        gaus (np.ndarray): gaussian function (normalized) of x
        params_gaus (dict): dictionary containing the input params
    '''
    if x is None:
        x = np.linspace(-sig*5, sig*5, int(sig*7), endpoint=True)

    gaus = 1/(np.sqrt(2*np.pi)*sig)*np.exp((-((x-mu)/sig) **2)/2)

    return gaus


def make_scaled_wave_basis(
    mother, 
    lens_waves, 
    lens_windows=None, 
    interp_kind='cubic', 
    fill_value=0,
):
    """
    Generates a set of wavelet-like basis functions by scaling a mother wavelet
    to different sizes. \n
    
    Note that this does not necessarily result in a true
    orthogonal 'wavelet' basis set. This function uses interpolation to adjust
    the mother wavelet's size, making it suitable for creating filter banks with
    various frequency resolutions.

    RH 2024

    Parameters:
    - mother (np.ndarray): 
        A 1D numpy array representing the mother wavelet used as the basis for scaling.
    - lens_waves (int, list, tuple, np.ndarray): 
        The lengths of the output waves. Can be a single integer or a list/array of integers.
    - lens_windows (int, list, tuple, np.ndarray, None): 
        The window lengths for each of the output waves. If None, defaults to
        the values in lens_waves. Can be a single integer (applied to all waves)
        or a list/array of integers corresponding to each wave length.
    - interp_kind (str): 
        Specifies the kind of interpolation as a string ('linear', 'nearest',
        'zero', 'slinear', 'quadratic', 'cubic', where 'zero', 'slinear',
        'quadratic' and 'cubic' refer to a spline interpolation of zeroth,
        first, second or third order).
    - fill_value (float): 
        Value used to fill in for requested points outside of the domain of the
        x_mother. Can be anything from scipy.interpolate.interp1d. If not
        provided, defaults to 0.

    Returns:
        (tuple):
        - waves (list): 
            List of the scaled wavelets.
        - xs (list):
            List of the x-values for each of the scaled wavelets.

    Example:
    ```
    mother_wave = np.cos(np.linspace(-2*np.pi, 2*np.pi, 10000, endpoint=True))
    lens_waves = [50, 100, 200]
    lens_windows = [100, 200, 400]
    waves, xs = make_scaled_wave_basis(mother_wave, lens_waves, lens_windows)
    ```
    """
    assert isinstance(mother, np.ndarray), "mother must be a 1D array"
    assert mother.ndim == 1, "mother must be a 1D array"
    
    arraylikes = (list, tuple, np.ndarray)
    if isinstance(lens_waves, arraylikes):
        lens_waves = np.array(lens_waves, dtype=int)
        if lens_windows is None:
            lens_windows = lens_waves
        if isinstance(lens_windows, int):
            lens_windows = np.array([lens_windows] * len(lens_waves), dtype=int)
        if isinstance(lens_windows, arraylikes):
            assert len(lens_waves) == len(lens_windows), "lens_waves and lens_windows must have the same length"
            lens_windows = np.array(lens_windows, dtype=int)
        else:
            raise ValueError("lens_windows must be an int or an array-like")
    elif isinstance(lens_waves, int):
        if lens_windows is None:
            lens_windows = lens_waves
        if isinstance(lens_windows, int):
            lens_waves = np.array([lens_waves], dtype=int)
            lens_windows = np.array([lens_windows], dtype=int)
        if isinstance(lens_windows, arraylikes):
            lens_waves = np.array([lens_waves] * len(lens_windows), dtype=int)
            lens_windows = np.array(lens_windows, dtype=int)
        else:
            raise ValueError("lens_windows must be an int or an array-like")
    else:
        raise ValueError("lens_waves must be an int or an array-like")


    x_mother = np.linspace(start=0, stop=1, num=len(mother), endpoint=True) - 0.5

    interpolator = scipy.interpolate.interp1d(
        x=x_mother,
        y=mother, 
        kind=interp_kind, 
        fill_value=fill_value, 
        bounds_error=False, 
        assume_sorted=True,
    )

    waves = []
    xs = []
    for i_wave, (l_wave, l_window) in enumerate(zip(lens_waves, lens_windows)):
        x_wave = (np.linspace(start=0, stop=1, num=l_window, endpoint=True) - 0.5) * (l_window / l_wave)
        wave = interpolator(x_wave)
        waves.append(wave)
        xs.append(x_wave)

    return waves, xs


def torch_hilbert(x, N=None, dim=0):
    """
    Computes the analytic signal using the Hilbert transform.
    Based on scipy.signal.hilbert
    RH 2022
    
    Args:
        x (nd tensor):
            Signal data. Must be real.
        N (int):
            Number of Fourier components to use.
            If None, then N = x.shape[dim]
        dim (int):
            Dimension along which to do the transformation.
    
    Returns:
        xa (nd tensor):
            Analytic signal of input x along dim
    """
    assert x.is_complex() == False, "x should be real"
    n = x.shape[dim] if N is None else N
    assert n >= 0, "N must be non-negative"

    xf = torch.fft.fft(input=x, n=n, dim=dim)
    m = torch.zeros(n, dtype=xf.dtype, device=xf.device)
    if n % 2 == 0: ## then even
        m[0] = m[n//2] = 1
        m[1:n//2] = 2
    else:
        m[0] = 1 ## then odd
        m[1:(n+1)//2] = 2

    if x.ndim > 1:
        ind = [np.newaxis] * x.ndim
        ind[dim] = slice(None)
        m = m[tuple(ind)]

    return torch.fft.ifft(xf * m, dim=dim)


def make_VQT_filters(    
    Fs_sample: int=1000,
    Q_lowF: int=3,
    Q_highF: int=20,
    F_min: int=10,
    F_max: int=400,
    n_freq_bins: int=55,
    win_size: Optional[int]=None,
    window_type: Union[str, np.ndarray, list, tuple]='gaussian',
    symmetry: str='center',
    taper_asymmetric: bool=True,
    mother_resolution: int=10000,
    plot_pref: bool=False,
):
    """
    Creates a set of filters for use in the VQT algorithm. \n
    Set Q_lowF and Q_highF to be the same value for a Constant Q Transform (CQT)
    filter set. Varying these values will varying the Q factor logarithmically
    across the frequency range. \n

    RH 2022

    Args:
        Fs_sample (float):
            Sampling frequency of the signal.
        Q_lowF (float):
            Q factor to use for the lowest frequency.
        Q_highF (float):
            Q factor to use for the highest frequency.
        F_min (float):
            Lowest frequency to use.
        F_max (float):
            Highest frequency to use (inclusive).
        n_freq_bins (int):
            Number of frequency bins to use.
        win_size (int, None):
            Size of the window to use, in samples. \n
            If None, will be set to the next odd number after Q_lowF * (Fs_sample / F_min).
        window_type (str, np.ndarray, list, tuple):
            Window type to use. \n
                * If string: Will be passed to scipy.signal.windows.get_window.
                  See that documentation for options. Except for 'gaussian',
                  which you should just pass the string 'gaussian' without any
                  other arguments.
                * If array-like: Will directly be used as the window for the low
                  frequency wavelet and will be scaled for higher frequencies.
        symmetry (str):
            Whether to use a symmetric window or a single-sided window.
            - 'center': Use a symmetric / centered / 'two-sided' window.
            - 'left': Use a one-sided, left-half window. Only left half of the
            filter will be nonzero.
            - 'right': Use a one-sided, right-half window. Only right half of the
            filter will be nonzero.
        taper_asymmetric (bool):
            Only used if symmetry is not 'center'. \n
            Whether to taper the center of the window by multiplying center
            sample of window by 0.5.
        mother_resolution (int):
            Resolution of the mother wavelet. Should be a large integer.
        plot_pref (bool):
            Whether to plot the filters.

    Returns:
        filters (Torch ndarray):
            Array of complex sinusoid filters.
            shape: (n_freq_bins, win_size)
        freqs (Torch array):
            Array of frequencies corresponding to the filters.
        wins (Torch ndarray):
            Array of window functions corresponding to each filter. \n
            shape: (n_freq_bins, win_size)
    """

    # if win_size % 2 != 1:
    #     raise ValueError("RH Error: win_size should be an odd integer")
    
    ## Make frequencies. Use a geometric spacing.
    freqs = np.geomspace(
        start=F_min,
        stop=F_max,
        num=n_freq_bins,
        endpoint=True,
        dtype=np.float32,
    )

    periods = 1 / freqs
    periods_inSamples = Fs_sample * periods

    if win_size is None:
        win_size = int(np.ceil(Q_lowF * (Fs_sample / F_min)))
        ## Make sure win_size is odd
        if win_size % 2 != 1:
            win_size += 1
        win_size = 3 if win_size < 3 else win_size

    ## Make windows
    if isinstance(window_type, str):
        ## Handle gaussian windows separately
        scales = np.geomspace(
            start=Q_lowF,
            stop=Q_highF,
            num=n_freq_bins,
            endpoint=True,
            dtype=np.float32,
        ) * periods_inSamples
        scales = np.clip(scales, a_min=1, a_max=None)
        if window_type == 'gaussian':
            ## Make sigmas for gaussian windows. Use a geometric spacing.
            window_type = ('gaussian', mother_resolution * 0.15)
        # else:
        ### Make mother wave
        mother_wave = scipy.signal.windows.get_window(window=window_type, Nx=mother_resolution, fftbins=False)
        
        wins, xs = make_scaled_wave_basis(mother_wave, lens_waves=scales, lens_windows=win_size)
        wins = torch.as_tensor(np.stack(wins, axis=0), dtype=torch.float32)

    elif isinstance(window_type, (np.ndarray, list, tuple)):
        mother_wave = np.array(window_type, dtype=np.float32)
    else:
        raise ValueError("window_type must be a string or an array-like")

        
    ### Make windows symmetric or asymmetric
    if symmetry=='center':
        pass
    else:
        heaviside = (torch.arange(win_size) <= win_size//2).float()
        if symmetry=='left':
            pass
        elif symmetry=='right':
            heaviside = torch.flip(heaviside, dims=[0])
        else:
            raise ValueError("symmetry must be 'center', 'left', or 'right'")
        wins *= heaviside
        ### Taper the center of the window by multiplying center sample of window by 0.5
        if taper_asymmetric:
            wins[:, win_size//2] = wins[:, win_size//2] * 0.5

    filts = torch.stack([torch.cos(torch.linspace(-np.pi, np.pi, win_size) * freq * (win_size/Fs_sample)) * win for freq, win in zip(freqs, wins)], dim=0)
    filts_complex = torch_hilbert(filts.T, dim=0).T
    ## Normalize filters to have unit magnitude
    filts_complex = filts_complex / torch.linalg.norm(filts_complex, ord=1, dim=1, keepdim=True)  ## Note: ord=1 is L1 norm. This makes the filters have unit magnitude.
    
    freqs = torch.as_tensor(freqs, dtype=torch.float32)

    ## Plot
    if plot_pref:
        plt.figure()
        plt.plot(freqs)
        plt.xlabel('filter num')
        plt.ylabel('frequency (Hz)')

        plt.figure()
        plt.imshow(wins / torch.max(wins, 1, keepdims=True)[0], aspect='auto')
        plt.ylabel('filter num')
        plt.title('windows (gaussian)')

        plt.figure()
        plt.plot(scales)
        plt.xlabel('filter num')
        plt.ylabel('window width scales')    

        plt.figure()
        plt.imshow(torch.real(filts_complex) / torch.max(torch.real(filts_complex), 1, keepdims=True)[0], aspect='auto', cmap='bwr', vmin=-1, vmax=1)
        plt.ylabel('filter num')
        plt.title('filters (real component)')


        worN=win_size*4
        filts_freq = np.array([scipy.signal.freqz(
            b=filt,
            fs=Fs_sample,
            worN=worN,
        )[1] for filt in filts_complex])

        filts_freq_xAxis = scipy.signal.freqz(
            b=filts_complex[0],
            worN=worN,
            fs=Fs_sample
        )[0]

        plt.figure()
        plt.plot(filts_freq_xAxis, np.abs(filts_freq.T));
        plt.xscale('log')
        plt.xlabel('frequency (Hz)')
        plt.ylabel('magnitude')

    return filts_complex, freqs, wins


from typing import Union, List, Tuple, Optional, Dict, Any, Sequence, Iterable, Type, Callable
import math

import torch
# import scipy.signal
from tqdm import tqdm
# import scipy.fftpack

from . import helpers

class VQT(torch.nn.Module):
    """
    Variable Q Transform. Class for applying the variable Q transform to
    signals. \n

    This function works differently than the VQT from librosa or nnAudio.
    This one does not use iterative lowpass filtering. \n 
    If fft_conv is False, then it uses a fixed set of filters, a Hilbert
    transform to compute the analytic signal, and then takes the magnitude. \n 
    If fft_conv is True, then it uses FFT convolution to compute the transform.
    \n
    
    Uses Pytorch for GPU acceleration, and allows gradients to pass through. \n

    Q: quality factor; roughly corresponds to the number of cycles in a
    filter. Here, Q is similar to the number of cycles within 4 sigma (95%)
    of a gaussian window. \n

    For running batches on GPU, transferring back to CPU tends to be the slowest
    part. \n

    RH 2022-2024

    Args:
        Fs_sample (float):
            Sampling frequency of the signal.
        Q_lowF (float):
            Q factor to use for the lowest frequency.
        Q_highF (float):
            Q factor to use for the highest frequency.
        F_min (float):
            Lowest frequency to use.
        F_max (float):
            Highest frequency to use.
        n_freq_bins (int):
            Number of frequency bins to use.
        win_size (int, None):
            Size of the window to use, in samples. \n
            If None, will be set to the next odd number after Q_lowF * (Fs_sample / F_min).
        window_type (str, np.ndarray, list, tuple):
            Window to use for the mother wavelet. \n
                * If string: Will be passed to scipy.signal.windows.get_window.
                See that documentation for options. Except for 'gaussian',
                which you should just pass the string 'gaussian' without any
                other arguments.
                * If array-like: Will be used as the window directly.
        symmetry (str):
            Whether to use a symmetric window or a single-sided window. \n
                * 'center': Use a symmetric / centered / 'two-sided' window.
                    \n
                * 'left': Use a one-sided, left-half window. Only left half
                of the filter will be nonzero. \n * 'right': Use a
                one-sided, right-half window. Only right half of the filter
                will be nonzero. \n
        taper_asymmetric (bool):
            Only used if symmetry is not 'center'. Whether to taper the
            center of the window by multiplying center sample of window by
            0.5.
        downsample_factor (int):
            Factor to downsample the signal by. If the length of the input
            signal is not divisible by downsample_factor, the signal will be
            zero-padded at the end so that it is.
        padding (str):
            Padding mode to use: \n
                * If fft_conv==False: ['valid', 'same'] \n
                * If fft_conv==True: ['full', 'valid', 'same'] \n
        fft_conv (bool):
            Whether to use FFT convolution. This is faster, but may be less
            accurate. If False, uses torch's conv1d.
        fast_length (bool):
            Whether to use scipy.fftpack.next_fast_len to 
                find the next fast length for the FFT.
            This may be faster, but uses more memory.
        take_abs (bool):
            Whether to return the complex version of the transform. If
            True, then returns the absolute value (envelope) of the
            transform. If False, returns the complex transform.
        filters (Torch tensor):
            Filters to use. If None, will make new filters. Should be
            complex sinusoids. shape: (n_freq_bins, win_size)
        verbose (int):
            Verbosity. True to print warnings.
        plot_pref (bool):
            Whether to plot the filters.
    """
    def __init__(
        self,
        Fs_sample: Union[int, float]=1000,
        Q_lowF: Union[int, float]=1,
        Q_highF: Union[int, float]=20,
        F_min: Union[int, float]=1,
        F_max: Union[int, float]=400,
        n_freq_bins: int=50,
        win_size: Optional[int]=None,
        window_type: Union[str, torch.Tensor]='gaussian',
        symmetry: str='center',
        taper_asymmetric: bool=True,
        downsample_factor: int=4,
        padding: str='same',
        fft_conv: bool=True,
        fast_length: bool=True,
        take_abs: bool=True,
        filters: Optional[torch.Tensor]=None,
        verbose: Union[int, bool]=True,
        plot_pref: bool=False,
    ):
        super().__init__()
        ## Prepare filters
        self.using_custom_filters = True if filters is not None else False
        self.filters = filters ## This line here is just for torch.jit.script to work. Delete it if you want to forget about jit.
        if filters is not None:
            ## Use provided filters
            self.filters = filters
        else:
            ## Make new filters
            self.filters, self.freqs, self.wins = helpers.make_VQT_filters(
                Fs_sample=Fs_sample,
                Q_lowF=Q_lowF,
                Q_highF=Q_highF,
                F_min=F_min,
                F_max=F_max,
                n_freq_bins=n_freq_bins,
                win_size=win_size,
                window_type=window_type,
                symmetry=symmetry,
                taper_asymmetric=taper_asymmetric,
                plot_pref=plot_pref,
            )
        
        ## Get win_size from filters
        win_size = self.filters.shape[-1]

        ## Make filters the parameters of the model
        self.filters = torch.nn.Parameter(self.filters, requires_grad=False)

        ## Gather parameters from arguments
        (
            self.Fs_sample, 
            self.Q_lowF, 
            self.Q_highF, 
            self.F_min, 
            self.F_max, 
            self.n_freq_bins, 
            self.win_size, 
            self.downsample_factor, 
            self.padding, 
            self.fft_conv,
            self.fast_length,
            self.take_abs, 
            self.plot_pref, 
         ) = (
            Fs_sample, 
            Q_lowF, 
            Q_highF, 
            F_min, 
            F_max, 
            n_freq_bins, 
            win_size, 
            downsample_factor, 
            padding, 
            fft_conv,
            fast_length,
            take_abs, 
            plot_pref, 
         )
        
        ## Warnings
        if verbose >= 1:
            ## Warn if win_size is even
            if win_size % 2 != 1:
                print("Warning: win_size is even. This will result in a non-centered window. The x_axis will be offset by 0.5. It is recommended to use an odd win_size.")
            ## Warn if win_size is > 1024 to use fft_conv
            if win_size > 1024 and fft_conv == False:
                print(f"Warning: win_size is {win_size}, which is large for conv1d. Consider using fft_conv=True for faster computation.")
            
    def forward(
        self,
        X: torch.Tensor,
    ) -> torch.Tensor:
        """
        Forward pass of VQT.

        Args:
            X (Torch tensor):
                Input signal.
                shape: (n_channels, n_samples)

        Returns:
            Spectrogram (Torch tensor):
                Spectrogram of the input signal.
                shape: (n_channels, n_samples_ds, n_freq_bins)
        """
        assert isinstance(X, torch.Tensor), "X should be a torch tensor"
        X = X.type(torch.float32)

        ## Check that X and filters are on the same device
        assert X.device == self.filters.device, "X and filters should be on the same device"

        if X.ndim==1:
            X = X[None,:]

        assert X.ndim==2, "X should be 2D"  ## (n_channels, n_samples)
        assert self.filters.ndim==2, "Filters should be 2D" ## (n_freq_bins, win_size)

        ## Make spectrograms
        specs = downsample(
            X=convolve(
                arr=X, 
                kernels=self.filters, 
                take_abs=self.take_abs,
                fft_conv=self.fft_conv,
                padding=self.padding,
                fast_length=self.fast_length,
            ), 
            ds_factor=self.downsample_factor,
        )

        return specs
        
    def get_freqs(self) -> torch.Tensor:
        """
        Get the frequencies of the spectrogram.
        
        Args:
            None

        Returns:
            torch.Tensor:
                Frequencies of the spectrogram. \n
                shape: (n_freq_bins,)
        """
        assert hasattr(self, 'freqs'), "freqs not found. This should not happen."
        return self.freqs
    
    def get_xAxis(self, n_samples: int) -> torch.Tensor:
        """
        Get the x-axis for the spectrogram. \n
        RH 2024

        Args:
            n_samples (int):
                Number of samples in the signal.

        Returns:
            torch.Tensor:
                x-axis for the spectrogram in units of samples. \n
                shape: (n_samples_ds,)
        """
        ## Make x_axis
        x_axis = make_conv_xAxis(
            n_s=n_samples,
            n_k=self.filters.shape[-1],
            padding=self.padding,
            downsample_factor=self.downsample_factor,
            device='cpu',
        )
        return x_axis

    def __repr__(self):
        if self.using_custom_filters:
            return f"VQT with custom filters"
        else:
            attributes_to_print = []
            for k, v in self.__dict__.items():
                if (k not in ['filters', 'freqs', 'wins']) and (not k.startswith('_')) and (not callable(v)):
                    attributes_to_print.append(k)
            return f"VQT object with parameters: {''.join([f'{k}={getattr(self, k)}, ' for k in attributes_to_print])[:-2]}"
        

def downsample(
    X: torch.Tensor, 
    ds_factor: int=4, 
) -> torch.Tensor:
    """
    Downsample a signal using average pooling. \n
    If X is complex, it will be split into magnitude and phase, downsampled,
    and then recombined. \n

    RH 2024

    Args:
        X (torch.Tensor):
            Signal to downsample. \n
            ``shape``: (..., n_samples)
        ds_factor (int):
            Factor to downsample the signal by.

    Returns:
        torch.Tensor:
            Downsampled signal.
    """
    if ds_factor == 1:
        return X
    
    ## Assert X is a torch tensor
    assert isinstance(X, torch.Tensor), "X should be a torch tensor"
    ## Ensure X.ndim in [2, 3]
    if X.ndim==1:
        X = X[None,:]
    assert X.ndim in [2, 3], "X should be 2D or 3D"  ## (n_channels, n_samples)
    
    ## Check is X is complex
    if X.is_complex() == False:
        return _helper_ds(X, ds_factor=ds_factor)
    elif X.is_complex() == True:
        ## Unfortunately, torch.nn.functional.avg_pool1d does not support complex numbers. So we have to split it up into
        ##  phases and magnitudes (convert imaginary to polar, split, downsample, recombine with polar to complex conversion)
        X = _helper_imag_to_polarReal(X)
        X = torch.stack([_helper_ds(X[ii], ds_factor=ds_factor) for ii in range(2)], dim=0)
        X = _helper_polarReal_to_imag(X)
        return X

    else:
        raise ValueError("X should be a torch tensor of type float or complex")
def _helper_polarReal_to_imag(arr: torch.Tensor) -> torch.Tensor:
    return arr[0] * torch.exp(1j * arr[1])
def _helper_imag_to_polarReal(arr: torch.Tensor) -> torch.Tensor:
    return torch.stack([torch.abs(arr), torch.angle(arr)], dim=0)
def _helper_ds(arr: torch.Tensor, ds_factor: int) -> torch.Tensor:
    return torch.nn.functional.avg_pool1d(
        arr,
        kernel_size=[int(ds_factor)],
        stride=ds_factor,
        ceil_mode=True,
        # padding=0,
        count_include_pad=False,  ## Prevents edge effects
    )

def convolve(
    arr: torch.Tensor, 
    kernels: torch.Tensor, 
    take_abs: bool=False,
    padding: str='same', 
    fft_conv: bool=False, 
    fast_length: bool=False,
) -> torch.Tensor:
    """
    Convolve a signal with a set of kernels. \n

    RH 2024

    Args:
        arr (torch.Tensor):
            Signal to convolve. \n
            ``shape``: (n_channels, n_samples)
        kernels (torch.Tensor):
            Kernels to convolve with. \n
            ``shape``: (n_kernels, win_size)
        take_abs (bool):
            Whether to take the absolute value of the result.
        padding (str):
            Padding mode to use: \n
                * If fft_conv==False: ['valid', 'same'] \n
                * If fft_conv==True: ['full', 'valid', 'same'] \n
        fft_conv (bool):
            Whether to use FFT convolution.
        fast_length (bool):
            Whether to use scipy.fftpack.next_fast_len to find the next fast
            length for the FFT.

    Returns:
        torch.Tensor:
            Result of the convolution. \n
            ``shape``: (n_channels, n_samples, n_kernels)
    """
    assert all(isinstance(arg, torch.Tensor) for arg in [arr, kernels]), "arr and kernels should be torch tensors"

    arr = arr[None,:] if arr.ndim==1 else arr
    kernels = kernels[None,:] if kernels.ndim==1 else kernels

    arr = arr[:,None,:]  ## Shape: (n_channels, 1, n_samples)
    # kernels = kernels  ## Shape: (n_kernels, win_size)
    
    if fft_conv:
        out = fftconvolve(
            x=arr,  
            y=(kernels)[None,:,:], 
            mode=padding,
            fast_length=fast_length,
        )
    else:
        flag_kernels_complex = kernels.is_complex()
        kernels = torch.flip(kernels, dims=[-1,])[:,None,:]  ## Flip because torch's conv1d uses cross-correlation, not convolution.
        
        if flag_kernels_complex:
            kernels_list = [torch.real(kernels), torch.imag(kernels)]
        else:
            kernels_list = [kernels,]

        out_conv = [torch.nn.functional.conv1d(
            input=arr, 
            weight=k, 
            padding=padding,
        ) for k in kernels_list]
        
        if flag_kernels_complex:
            out = torch.complex(out_conv[0], out_conv[1])
        else:
            out = out_conv[0]
        
    if take_abs:
        out = torch.abs(out)
    return out


def fftconvolve(
    x: torch.Tensor, 
    y: torch.Tensor, 
    mode: str='valid',
    fast_length: bool=False,
) -> torch.Tensor:
    """
    Convolution using the FFT method. \n
    This is adapted from of torchaudio.functional.fftconvolve that handles
    complex numbers. Code is added for handling complex inputs. \n
    NOTE: For mode='same' and y length even, torch's conv1d convention is used,
    which pads 1 more at the end and 1 fewer at the beginning (which is
    different from numpy/scipy's convolve). See apply_padding_mode for more
    details. \n

    RH 2024

    Args:
        x (torch.Tensor):
            First input. (signal) \n
            Convolution performed along the last dimension.
        y (torch.Tensor):
            Second input. (kernel) \n
            Convolution performed along the last dimension.
        mode (str):
            Padding mode to use. ['full', 'valid', 'same']
        fast_length (bool):
            Whether to use scipy.fftpack.next_fast_len to 
             find the next fast length for the FFT.
            Set to False if you want to use backpropagation.

    Returns:
        torch.Tensor:
            Result of the convolution. \n
            Padding applied to last dimension. \n
            ``shape``: (..., n_samples)
    """
    ## Compute the convolution
    n_original = x.shape[-1] + y.shape[-1] - 1
    # n = scipy.fftpack.next_fast_len(n_original) if fast_length else n_original
    n = next_fast_len(n_original) if fast_length else n_original
    # n = n_original
    f = torch.fft.fft(x, n=n, dim=-1) * torch.fft.fft(y, n=n, dim=-1)
    fftconv_xy = torch.fft.ifft(f, n=n, dim=-1)
    return apply_padding_mode(
        conv_result=fftconv_xy,
        x_length=x.shape[-1],
        y_length=y.shape[-1],
        mode=mode,
    )

## For some reason jit is slower here
# @torch.jit.script
def next_fast_len(size: int) -> int:
    """
    Taken from PyTorch Forecasting:
    Returns the next largest number ``n >= size`` whose prime factors are all
    2, 3, or 5. These sizes are efficient for fast fourier transforms.
    Equivalent to :func:`scipy.fftpack.next_fast_len`.

    Implementation from pyro

    :param int size: A positive number.
    :returns: A possibly larger number.
    :rtype int:
    """
    assert isinstance(size, int) and size > 0
    next_size = size
    while True:
        remaining = next_size
        for n in (2, 3, 5):
            while remaining % n == 0:
                remaining = remaining // n
        if remaining == 1:
            return next_size
        next_size += 1


def apply_padding_mode(
    conv_result: torch.Tensor, 
    x_length: int, 
    y_length: int, 
    mode: str = "valid",
) -> torch.Tensor:
    """
    This is adapted from torchaudio.functional._apply_convolve_mode. \n
    NOTE: This function has a slight change relative to torchaudio's version.
    For mode='same', ceil rounding is used. This results in fftconv matching the
    result of conv1d. However, this then results in it not matching the result of
    scipy.signal.fftconvolve. This is a tradeoff. The difference is only a shift
    in 1 sample when y_length is even. This phenomenon is a result of how conv1d
    handles padding, and the fact that conv1d is actually cross-correlation, not
    convolution. \n

    RH 2024

    Args:
        conv_result (torch.Tensor):
            Result of the convolution.
            Padding applied to last dimension.
        x_length (int):
            Length of the first input.
        y_length (int):
            Length of the second input.
        mode (str):
            Padding mode to use.

    Returns:
        torch.Tensor:
            Result of the convolution with the specified padding mode. \n
            ``shape``: (..., n_samples)
    """
    n = x_length + y_length - 1
    valid_convolve_modes = ["full", "valid", "same"]
    if mode == "full":
        return conv_result
    elif mode == "valid":
        len_target = max(x_length, y_length) - min(x_length, y_length) + 1
        idx_start = (n - len_target) // 2
        return conv_result[..., idx_start : idx_start + len_target]
    elif mode == "same":
        # idx_start = (conv_result.size(-1) - x_length) // 2  ## This is the original line from torchaudio
        idx_start = math.ceil((n - x_length) / 2)  ## This line is different from torchaudio
        return conv_result[..., idx_start : idx_start + x_length]
    else:
        raise ValueError(f"Unrecognized mode value '{mode}'. Please specify one of {valid_convolve_modes}.")


def make_conv_xAxis(
    n_s: int,
    n_k: int,
    padding: str='same',
    downsample_factor: int=4,
    device: torch.device='cpu',
) -> torch.Tensor:
    """
    Make the x-axis for the result of a convolution.
    This is adapted from torchaudio.functional._make_conv_xAxis.

    RH 2024

    Args:
        n_s (int):
            Length of the signal.
        n_k (int):
            Length of the kernel.
        padding (str):
            Padding mode to use.
        downsample_factor (int):
            Factor to downsample the signal by.
        device (str):
            Device to use.

    Returns:
        torch.Tensor:
            x-axis for the result of a convolution. \n
            ``shape``: (n_samples_ds,)
    """

    ## If n_k is odd, then no offset, if even, then offset by 0.5
    ### PyTorch's conv1d and for 'same' pads more to the right, so on the first index of the output the kernel is centered at an offset of 0.5
    offset = 0.5 if n_k % 2 == 0 else 0.0

    x_axis_full = torch.arange(
        -(n_k-1)//2 + offset,
        n_s + (n_k-1)//2 + offset,
        dtype=torch.float32,
        device=device,
    )
    ### Then, apply padding mode to it
    x_axis_padModed = apply_padding_mode(
        conv_result=x_axis_full,
        x_length=n_s,
        y_length=n_k,
        mode=padding,
    ).squeeze()
    ### Then, let's downsample it
    x_axis = downsample(
        X=x_axis_padModed[None,None,:],
        ds_factor=downsample_factor,
    ).squeeze().to(device)

    return x_axis